{$R-,S-,I-,X+}

{*********************************************************}
{*                   COMSPY16.PAS 1.01                   *}
{*        Copyright (c) TurboPower Software 1993         *}
{*                 All rights reserved.                  *}
{*********************************************************}

program ComSpy16;
  {-Spy on USER.EXE communication functions}

uses
  {-----RTL}
  WinTypes,
  WinProcs,
  Strings,
  {$IFDEF VER80}
  Messages,
  {$ENDIF}
  WinCrt,
  {-----CommSpy}
  CSImport;

{$I COMSPY16.INC}             {Shared declarations}
{$R COMSPY16.RES}             {Options dialog box}

const
  {Option dialog box constants}
  id_SpyNone     = 101;
  id_SpyTotal    = 102;
  id_SpyTrace    = 103;
  id_SpyCapture  = 104;
  id_FileLog     = 105;
  id_FileName    = 106;
  id_SuppressGCE = 107;

  {Miscellaneous}
  CRLF = #13#10;

var
  CrtWinProc     : TFarProc;    {WinCrt's window proc}
  CrtWindow      : hWnd;        {WinCrt's window handle}
  Logging        : Boolean;     {True when logging to a file}
  NewLogging     : Boolean;     {Newly requested logging value}
  LogFileName    : array[0..80] of Char;  {Log file name}
  LogF           : Text;        {Log file}
  SpyMode        : Byte;        {Current spying mode}
  NewSpyMode     : Byte;        {New requested spying mode}
  SuppressGCE    : Boolean;     {True to suppress GetCommError logging}

type
  {Used to decide when to issue new Transmit/Receive headings}
  TDirection = (dNone, dTransmit, dReceive);

const
  {Miscellaneous}
  LastDirection  : TDirection = dNone;
  cs_Options     = 101;
  Digits         : array[0..$F] of Char = '0123456789ABCDEF';

{General purpose routines}

  function HexB(Dest : PChar; B : Byte) : PChar;
    {-Return hex string for byte}
  begin
    HexB := Dest;
    Dest^ := Digits[B shr 4];
    Inc(Dest);
    Dest^ := Digits[B and $F];
    Inc(Dest);
    Dest^ := #0;
  end;

  function FormatChar(P : PChar; C : Char) : PChar;
    {-Return a formatting string, 'A' for normal chars, '[XX]' for ctl chars}
  var
    P1 : array[0..2] of Char;
  begin
    case C of
      #0..#32,
      #127..#255 :
        begin
          HexB(P1, Byte(C));
          StrCopy(P, '[');
          StrCat(P, P1);
          StrCat(P, ']');
        end;
      else
        P[0] := C;
        P[1] := #0;
    end;
    FormatChar := P;
  end;

  procedure SafeYield;
    {-Yield to other applications}
  var
    Msg : TMsg;
  begin
    if PeekMessage(Msg, 0, 0, 0, PM_REMOVE) then begin
      TranslateMessage(Msg);
      DispatchMessage(Msg);
    end;
  end;

{Logging functions}

  procedure StartLogging;
    {-Create and open a log file}
  begin
    Assign(LogF, LogFileName);
    Rewrite(LogF);
    if IoResult <> 0 then
      Logging := False;
  end;

  procedure StopLogging;
    {-Close the current log file}
  begin
    Close(LogF);
    if IoResult <> 0 then ;
  end;

{Optios dialog box}

  function csDlgProc(hWndDlg : HWnd; Msg, wParam : word; lParam : LongInt)  : Word; far;
    {-Dialog box procedure}
  begin
    case Msg of
      wm_InitDialog :
        begin
          SetDlgItemText(hWndDlg, id_FileName, LogFileName);
          case SpyMode of
            smNone    : CheckDlgButton(hWndDlg, id_SpyNone, 1);
            smTotal   : CheckDlgButton(hWndDlg, id_SpyTotal, 1);
            smTrace   : CheckDlgButton(hWndDlg, id_SpyTrace, 1);
            smCapture : CheckDlgButton(hWndDlg, id_SpyCapture, 1);
          end;
          if Logging then
            CheckDlgButton(hWndDlg, id_FileLog, 1);
          if SuppressGCE then
            CheckDlgButton(hWndDlg, id_SuppressGCE, 1);
        end;

      wm_Command    :
        case wParam of
          id_ok :
            begin
              GetDlgItemText(hWndDlg, id_FileName, LogFileName, 80);
              NewLogging := (IsDlgButtonChecked(hWndDlg, id_FileLog) <> 0);
              if (IsDlgButtonChecked(hWndDlg, id_SpyTotal) <> 0) then
                NewSpyMode := smTotal
              else if (IsDlgButtonChecked(hWndDlg, id_SpyTrace) <> 0) then
                NewSpyMode := smTrace
              else if (IsDlgButtonChecked(hWndDlg, id_SpyCapture) <> 0) then
                NewSpyMode := smCapture
              else
                NewSpyMode := smNone;
              SuppressGCE := (IsDlgButtonChecked(hWndDlg, id_SuppressGCE) <> 0);
              EndDialog(hWndDlg, 1);
            end;
          id_Cancel :
            EndDialog(hWndDlg, 0);
        end;
      else
        csDlgProc := 0;
    end;
  end;

  procedure ExecOptionsDialog;
    {-Execute the options dialog box}
  var
    Result : Integer;
  begin
    Result := DialogBox(HInstance, 'CSOPTIONS', CrtWindow, @csDlgProc);
    if Result = 1 then begin
      if NewSpyMode <> SpyMode then begin
        if SpyMode <> smNone then
          StopSpying(CrtWindow);
        if NewSpyMode <> smNone then
          StartSpying(CrtWindow, NewSpyMode);
        SpyMode := NewSpyMode;
      end;
      if NewLogging <> Logging then begin
        if Logging then
          StopLogging
        else
          StartLogging;
        Logging := NewLogging;
      end;
    end;
  end;

{Replacement window procedure for CrtWindow}

  function CommSpyWinProc(HW : hWnd; Msg, wParam : Word;
                          lParam : Longint) : Longint; export;
    {-lParam points to data, wParam has size}
  type
    Chars = array[1..65521] of Char;
  const
    FirstCall : Boolean = True;
    MaxCol = 70;
    Col : Word = 0;
  var
    Menu : hMenu;

    procedure ShowTrace;
      {-APRO style trace of data pointed to by lParam}
    var
      I : Word;
      FC : array[0..4] of Char;
    begin
      {Show received data}
      for I := 1 to wParam do begin
        {Display a formatted character}
        FormatChar(FC, Chars(Pointer(lParam)^)[I]);
        WriteBuf(FC, StrLen(FC));

        if Logging then
          Write(LogF, FC);

        {Handle line wrapping}
        Inc(Col, StrLen(FC));
        if Col > MaxCol then begin
          WriteBuf(CRLF, 2);
          if Logging then
            WriteLn(LogF);
          Col := 0;
        end;
      end;
    end;

    procedure ShowCapture;
      {-Capture-type display of data pointed to by lParam}
    var
      I : Word;
      FC : array[0..4] of Char;
      C : Char;
    begin
      {Show received data}
      for I := 1 to wParam do begin
        {Display a formatted character}
        C := Chars(Pointer(lParam)^)[I];
        WriteChar(C);

        if Logging then
          Write(LogF, C);

        {Handle line wrapping}
        if C = #13 then
          Col := 0
        else begin
          Inc(Col);
          if Col > MaxCol then begin
            WriteBuf(CRLF, 2);
            if Logging then
              WriteLn(LogF);
            Col := 0;
          end;
        end;
      end;
    end;

    procedure ShowHeader(NewDirection : TDirection);
      {-Write new header if necessary}
    begin
      if LastDirection <> NewDirection then begin
        LastDirection := NewDirection;
        Col := 0;

        {Send to CRT window}
        case NewDirection of
          dTransmit : WriteBuf(CRLF+CRLF+'Transmit:'+CRLF, 15);
          dReceive  : WriteBuf(CRLF+CRLF+'Receive:'+CRLF, 13);
        end;

        {Send to log file if necessary}
        if Logging then
          case NewDirection of
            dTransmit : WriteLn(LogF, CRLF+CRLF+'Transmit:');
            dReceive  : WriteLn(LogF, CRLF+CRLF+'Receive: ');
          end;
      end;
    end;

  begin
    case Msg of
      wm_InitMenu :
        if FirstCall then begin
          {Change system menu}
          FirstCall := False;
          Menu := GetSystemMenu(HW, False);
          InsertMenu(Menu, 5, mf_ByPosition+mf_Separator, 0, nil);
          InsertMenu(Menu, 6, mf_ByPosition+mf_String, cs_Options, '&Spy options');
          CommSpyWinProc := CallWindowProc(CrtWinProc, HW, Msg, wParam, lParam);
        end else
          CommSpyWinProc := CallWindowProc(CrtWinProc, HW, Msg, wParam, lParam);

      wm_SysCommand :
        if wParam = cs_Options then begin
          ExecOptionsDialog;
          CommSpyWinProc := 0;
        end else
          CommSpyWinProc := CallWindowProc(CrtWinProc, HW, Msg, wParam, lParam);

      wm_Destroy :
        begin
          StopSpying(CrtWindow);
          CommSpyWinProc := CallWindowProc(CrtWinProc, HW, Msg, wParam, lParam);
        end;

      wm_csReadComm :
        begin
          case SpyMode of
            smTrace   :
              begin
                ShowHeader(dReceive);
                ShowTrace;
              end;
            smCapture :
              begin
                ShowHeader(dReceive);
                ShowCapture;
              end;
            smTotal   :
              begin
                WriteBuf(CRLF+'ReadComm: ', 12);
                ShowTrace;
              end;
          end;
          CommSpyWinProc := 0;
        end;

      wm_csWritecomm :
        begin
          case SpyMode of
            smTrace   :
              begin
                ShowHeader(dTransmit);
                ShowTrace;
              end;
            smCapture :
              begin
                ShowHeader(dTransmit);
                ShowCapture;
              end;
            smTotal   :
              begin
                WriteBuf(CRLF+'WriteComm: ', 13);
                if Logging then
                  Write(LogF, CRLF+'WriteComm: ');
                ShowTrace;
              end;
          end;
          CommSpyWinProc := 0;
        end;

      wm_csOpenComm :
        begin
          WriteBuf(CRLF+'OpenComm: ', 10);
          WriteBuf(Pointer(lParam), wParam);
          if Logging then
            Write(LogF, CRLF+'OpenComm: ', PChar(lParam));
        end;

      wm_csBuildCommDCB :
        begin
          WriteBuf(CRLF+'BuildCommDCB', 14);
          if Logging then
            Write(LogF, CRLF+'BuildCommDCB');
        end;

      wm_csClearCommBreak :
        begin
          WriteBuf(CRLF+'ClearCommBreak', 16);
          if Logging then
            Write(LogF, CRLF+'ClearCommBreak');
        end;

      wm_csCloseComm :
        begin
          WriteBuf(CRLF+'CloseComm', 10);
          if Logging then
            Write(LogF, CRLF+'CloseComm');
        end;

      wm_csEscapeCommFunction :
        begin
          WriteBuf(CRLF+'EscapeCommFunction: ', 20);
          WriteBuf(Pointer(lParam), wParam);
          if Logging then
            Write(LogF, CRLF+'EscapeCommFunction: ', PChar(lParam));
        end;

      wm_csFlushComm :
        begin
          WriteBuf(CRLF+'FlushComm: ', 11);
          WriteBuf(Pointer(lParam), wParam);
          if Logging then
            Write(LogF, CRLF+'FlushComm: ', PChar(lParam));
        end;

      wm_csGetCommError :
        if not SuppressGCE then begin
          WriteBuf(CRLF+'GetCommError: ', 16);
          WriteBuf(Pointer(lParam), wParam);
          if Logging then
            Write(LogF, CRLF+'GetCommError: ', PChar(lParam));
        end;

      wm_csGetCommEventMask :
        begin
          WriteBuf(CRLF+'GetCommEventMask: ', 20);
          WriteBuf(Pointer(lParam), wParam);
          if Logging then
            Write(LogF, CRLF+'GetCommEventMask: ', PChar(lParam));
        end;

      wm_csGetCommState :
        begin
          WriteBuf(CRLF+'GetCommState: ', 16);
          WriteBuf(Pointer(lParam), wParam);
          if Logging then
            Write(LogF, CRLF+'GetCommState: ', PChar(lParam));
        end;

      wm_csSetCommBreak :
        begin
          WriteBuf(CRLF+'SetCommBreak', 14);
          if Logging then
            Write(LogF, CRLF+'SetCommBreak');
        end;

      wm_csSetCommEventMask :
        begin
          WriteBuf(CRLF+'SetCommEventMask: ', 20);
          WriteBuf(Pointer(lParam), wParam);
          if Logging then
            Write(LogF, CRLF+'SetCommEventMask: ', PChar(lParam));
        end;

      wm_csSetCommState :
        begin
          WriteBuf(CRLF+'SetCommState: ', 16);
          WriteBuf(Pointer(lParam), wParam);
          if Logging then
            Write(LogF, CRLF+'SetCommState: ', PChar(lParam));
        end;

      wm_csTransmitCommChar :
        begin
          WriteBuf(CRLF+'TransmitCommChar: ', 20);
          WriteBuf(Pointer(lParam), wParam);
          if Logging then
            Write(LogF, CRLF+'TransmitCommChar: ', PChar(lParam));
        end;

      wm_csUngetCommChar :
        begin
          WriteBuf(CRLF+'UngetCommChar: ', 17);
          WriteBuf(Pointer(lParam), wParam);
          if Logging then
            Write(LogF, CRLF+'UngetCommChar: ', PChar(lParam));
        end;

      else
        {pass the message to the regular window procedure}
        CommSpyWinProc := CallWindowProc(CrtWinProc, HW, Msg, wParam, lParam);
    end;
  end;

begin
  {Set size of WinCrt's virtual screen}
  ScreenSize.X := 80;
  ScreenSize.Y := 65520 div 80;

  {Set window size to full width of display}
  WindowSize.X := GetSystemMetrics(sm_CXScreen);
  WindowSize.Y := 200;

  {Set window position to top of screen}
  WindowOrg.X := 0;
  WindowOrg.Y := 0;

  {Set window titles}
  StrCopy(WindowTitle, csTitle);
  InactiveTitle := csTitle;

  {Initialize WinCrt}
  InitWinCrt;

  {Find WinCrt's window}
  CrtWindow := FindWindow('TPWINCRT', csTitle);

  {Change address of window proc and save old value}
  CrtWinProc := TFarProc(SetWindowLong(CrtWindow, gwl_WndProc, Longint(@CommSpyWinProc)));

  {Start spying in smTrace mode, no logging, suppress GetCommError messages}
  SpyMode := smTrace;
  StartSpying(CrtWindow, SpyMode);
  Logging := False;
  SuppressGCE := True;

  {Set default log file name}
  StrCopy(LogFileName, 'COMSPY16.LOG');
end.
