{$R-,S-,I-,X+}

{*********************************************************}
{*                     CSDLL.PAS 1.01                    *}
{*        Copyright (c) TurboPower Software 1993         *}
{*                 All rights reserved.                  *}
{*********************************************************}

Library CSDLL;
  {-Patch USER.EXE communication functions}

uses
  {------RTL}
  WinTypes,
  WinProcs,
  {$IFNDEF VER80}
  Win31,
  {$ENDIF}
  Strings;

{$I COMSPY16.INC}             {Shared declarations}

const
  {Various limits}
  MaxMsgSize     = 8192;      {Max message size}
  MaxPatches     = 20;        {Max number of patched procedures}

  {Miscellaneous}
  MsgWin         : HWnd = 0;  {Window to get messages}
  SpyMode        : Byte = 0;  {Current spying mode}

type
  {For sending message data}
  PMsgBuf = ^TMsgBuf;
  TMsgBuf = array[1..MaxMsgSize] of Byte;

  {For storing patched and original code}
  CodeArray = array[1..5] of Byte;

  {For storing all information about a patch}
  TPatch = record
    OrigCode      : CodeArray;
    JmpCode       : CodeArray;
    PatchSelector : Word;
    PatchPoint    : Pointer;
    PatchSeg      : Word;
  end;
  TPatchArray = array[1..MaxPatches] of TPatch;

var
  Patch     : TPatchArray;       {List of patches}
  MsgBuf    : PMsgBuf;           {Buffer for sending data in message}

{Forward declarations needed for patching}

  function csReadComm(Cid : Integer; Buf : PChar; Size : Integer) : Integer; export; forward;
  function csWriteComm(Cid : Integer; Buf : PChar; Size : Integer) : Integer; export; forward;
  function csOpenComm(ComName: PChar; InQueue, OutQueue: Word): Integer; export; forward;
  function csBuildCommDCB(Def : PChar; var DCB : TDCB) : Integer; export; forward;
  function csClearCommBreak(Cid: Integer): Integer; export; forward;
  function csCloseComm(Cid: Integer): Integer; export; forward;
  function csEscapeCommFunction(Cid: Integer; Func: Integer): LongInt; export; forward;
  function csFlushComm(Cid, Queue: Integer): Integer; export; forward;
  function csGetCommError(Cid : Integer; var State : TComStat) : Integer; export; forward;
  function csGetCommEventMask(Cid, EvtMask: Integer): Word; export; forward;
  function csGetCommState(Cid: Integer; var DCB: TDCB): Integer; export; forward;
  function csSetCommBreak(Cid: Integer): Integer; export; forward;
  function csSetCommEventMask(Cid: Integer; EvtMask: Word): PWord; export; forward;
  function csSetCommState(var DCB: TDCB): Integer; export; forward;
  function csTransmitCommChar(Cid: Integer; Chr: Char): Integer; export; forward;
  function csUngetCommChar(Cid: Integer; Chr: Char): Integer; export; forward;

{General purpose routines}

  function Long2Str(Dest : PChar; L : LongInt) : PChar;
    {-Convert a long/word/integer/byte/shortint to a string}
  var
    S : string;
  begin
    Str(L, S);
    Long2Str := StrPCopy(Dest, S);
  end;

{Patching routines}

  procedure PatchWindowsProc(Index : Word; Old, New : Pointer);
    {-Patch Old address with JMP NEW, fill in Patch array}
  type
    QB = array[1..4] of Byte;
    OS = record
      O : Word;
      S : Word;
    end;
  var
    TempSelector : Word;
    I, J : Word;
    Result : Integer;
  begin
    with Patch[Index] do begin
      {Make selector for patch point}
      PatchSeg := OS(Old).S;
      I := GlobalPageLock(PatchSeg);
      PatchSelector := AllocSelector(PatchSeg);
      PrestoChangoSelector(PatchSeg, PatchSelector);
      PatchPoint := Ptr(PatchSelector, OS(Old).O);

      {Save original code}
      Move(PatchPoint^, OrigCode, 5);

      {Make a patch instruction}
      JmpCode[1] := $EA;
      J := 2;
      for I := 1 to 4 do begin
        JmpCode[J] := QB(New)[I];
        Inc(J);
      end;

      {Patch entry point}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

  procedure UnpatchWindowsProc(Index : Word);
    {-Remove patch Index from Patch array}
  var
    I : Word;
  begin
    with Patch[Index] do begin
      Move(OrigCode, PatchPoint^, 5);
      I := GlobalPageUnlock(PatchSeg);
      I := FreeSelector(PatchSelector);
    end;
  end;

  procedure PatchAll;
    {-Patch all entry points}
  begin
    PatchWindowsProc(1, @ReadComm, @csReadComm);
    PatchWindowsProc(2, @WriteComm, @csWriteComm);
    PatchWindowsProc(3, @OpenComm, @csOpenComm);
    PatchWindowsProc(4, @BuildCommDCB, @csBuildCommDCB);
    PatchWindowsProc(5, @ClearCommBreak, @csClearCommBreak);
    PatchWindowsProc(6, @CloseComm, @csCloseComm);
    PatchWindowsProc(7, @EscapeCommFunction, @csEscapeCommFunction);
    PatchWindowsProc(8, @FlushComm, @csFlushComm);
    PatchWindowsProc(9, @GetCommError, @csGetCommError);
    PatchWindowsProc(10, @GetCommEventMask, @csGetCommEventMask);
    PatchWindowsProc(11, @GetCommState, @csGetCommState);
    PatchWindowsProc(12, @SetCommBreak, @csSetCommBreak);
    PatchWindowsProc(13, @SetCommEventMask, @csSetCommEventMask);
    PatchWindowsProc(14, @SetCommState, @csSetCommState);
    PatchWindowsProc(15, @TransmitCommChar, @csTransmitCommChar);
    PatchWindowsProc(16, @UngetCommChar, @csUngetCommChar);
  end;

  procedure UnpatchAll;
    {-Unpatch all entry points}
  var
    I : Word;
  begin
    for I := 1 to 16 do
      UnpatchWindowsProc(I);
  end;

  procedure PatchReadWrite;
    {-Patch just read/write entry points}
  begin
    PatchWindowsProc(1, @ReadComm, @csReadComm);
    PatchWindowsProc(2, @WriteComm, @csWriteComm);
  end;

  procedure UnpatchReadWrite;
    {-Unpatch just read/write entry points}
  begin
    UnpatchWindowsProc(1);
    UnpatchWindowsProc(2);
  end;

{Patched procedures}

  function csReadComm(Cid : Integer; Buf : PChar; Size : Integer) : Integer;
    {-Unpatch ReadComm, call it, log it, repatch for next time}
  var
    fResult : Integer;
  begin
    {Unpatch ReadComm}
    with Patch[1] do begin
      {Remove patch}
      Move(OrigCode, PatchPoint^, 5);

      {Execute it}
      fResult := ReadComm(Cid, Buf, Size);
      csReadComm := fResult;

      {Fill message buf with received data}
      if fResult > 0 then
        Move(Buf^, MsgBuf^, fResult);

      {Log message}
      if fResult < 0 then
        fResult := 0;
      SendMessage(MsgWin, wm_csReadComm, fResult, LongInt(MsgBuf));

      {Repatch}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

  function csWriteComm(Cid : Integer; Buf : PChar; Size : Integer) : Integer;
    {-Unpatch ReadComm, call it, log it, repatch for next time}
  var
    fResult : Integer;
  begin
    with Patch[2] do begin
      Move(OrigCode, PatchPoint^, 5);

      {Execute it}
      fResult := WriteComm(Cid, Buf, Size);
      csWriteComm := fResult;

      {Fill message buf with transmitted data}
      if Size > 0 then
        Move(Buf^, MsgBuf^, Size);

      {Send results message}
      SendMessage(MsgWin, wm_csWriteComm, Size, LongInt(MsgBuf));

      {re-patch}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

  function csOpenComm(ComName: PChar; InQueue, OutQueue: Word): Integer;
    {-Unpatch, call it, log it, repatch for next time}
  var
    fResult : Integer;
    P : array[0..5] of Char;
  begin
    with Patch[3] do begin
      Move(OrigCode, PatchPoint^, 5);

      {Execute it}
      fResult := OpenComm(ComName, InQueue, OutQueue);
      csOpenComm := fResult;

      {Fill message buf with info about OpenComm}
      StrCopy(PChar(MsgBuf), ComName);
      StrCat(PChar(MsgBuf), ' ');
      StrCat(PChar(MsgBuf), Long2Str(P, InQueue));
      StrCat(PChar(MsgBuf), ' ');
      StrCat(PChar(MsgBuf), Long2Str(P, OutQueue));
      SendMessage(MsgWin, wm_csOpenComm, StrLen(PChar(MsgBuf)), LongInt(MsgBuf));

      {re-patch}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

  function csBuildCommDCB(Def : PChar; var DCB : TDCB) : Integer;
    {-Unpatch, call it, log it, repatch for next time}
  var
    fResult : Integer;
  begin
    with Patch[4] do begin
      Move(OrigCode, PatchPoint^, 5);

      {Execute it}
      fResult := BuildCommDCB(Def, DCB);
      csBuildCommDCB := fResult;

      {No data to send}
      SendMessage(MsgWin, wm_csBuildCommDCB, 0, 0);

      {re-patch}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

  function csClearCommBreak(Cid: Integer): Integer;
    {-Unpatch, call it, log it, repatch for next time}
  var
    fResult : Integer;
  begin
    with Patch[5] do begin
      Move(OrigCode, PatchPoint^, 5);

      {Execute it}
      fResult := ClearCommBreak(Cid);
      csClearCommBreak := fResult;

      {No data to send}
      SendMessage(MsgWin, wm_csClearCommBreak, 0, 0);

      {re-patch}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

  function csCloseComm(Cid: Integer): Integer;
    {-Unpatch, call it, log it, repatch for next time}
  var
    fResult : Integer;
  begin
    with Patch[6] do begin
      Move(OrigCode, PatchPoint^, 5);

      {Execute it}
      fResult := CloseComm(Cid);
      csCloseComm := fResult;

      {No data to send}
      SendMessage(MsgWin, wm_csCloseComm, 0, 0);

      {re-patch}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

  function csEscapeCommFunction(Cid: Integer; Func: Integer): LongInt;
    {-Unpatch, call it, log it, repatch for next time}
  var
    fResult : Integer;
    P : array[0..5] of Char;
  begin
    with Patch[7] do begin
      Move(OrigCode, PatchPoint^, 5);

      {Execute it}
      fResult := EscapeCommFunction(Cid, Func);
      csEscapeCommFunction := fResult;

      Long2Str(P, Func);
      StrCopy(PChar(MsgBuf), P);
      SendMessage(MsgWin, wm_csEscapeCommFunction, StrLen(P), LongInt(MsgBuf));

      {re-patch}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

  function csFlushComm(Cid, Queue: Integer): Integer;
    {-Unpatch, call it, log it, repatch for next time}
  var
    fResult : Integer;
  begin
    with Patch[8] do begin
      Move(OrigCode, PatchPoint^, 5);

      {Execute it}
      fResult := FlushComm(Cid, Queue);
      csFlushComm := fResult;

      if Queue = 0 then
        StrCopy(PChar(MsgBuf), ' output')
      else
        StrCopy(PChar(MsgBuf), ' input');
      SendMessage(MsgWin, wm_csFlushComm, StrLen(PChar(MsgBuf)), LongInt(MsgBuf));

      {re-patch}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

  function csGetCommError(Cid : Integer; var State : TComStat) : Integer;
    {-Unpatch, call it, log it, repatch for next time}
  var
    fResult : Integer;
  begin
    with Patch[9] do begin
      Move(OrigCode, PatchPoint^, 5);

      {Execute it}
      fResult := GetCommError(Cid, WinTypes.TComStat(State));
      csGetCommError := fResult;

      {Just retrun the result as a string for now}
      Long2Str(PChar(MsgBuf), fResult);
      SendMessage(MsgWin, wm_csGetCommError, StrLen(PChar(MsgBuf)), LongInt(MsgBuf));

      {re-patch}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

  function csGetCommEventMask(Cid, EvtMask: Integer): Word;
    {-Unpatch, call it, log it, repatch for next time}
  var
    fResult : Integer;
  begin
    with Patch[10] do begin
      Move(OrigCode, PatchPoint^, 5);

      {Execute it}
      fResult := GetCommEventMask(Cid, EvtMask);
      csGetCommEventMask := fResult;

      Long2Str(PChar(MsgBuf), EvtMask);
      SendMessage(MsgWin, wm_csGetCommEventMask, StrLen(PChar(MsgBuf)), LongInt(MsgBuf));

      {re-patch}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

  function csGetCommState(Cid: Integer; var DCB: TDCB): Integer;
    {-Unpatch, call it, log it, repatch for next time}
  var
    fResult : Integer;
  begin
    with Patch[11] do begin
      Move(OrigCode, PatchPoint^, 5);

      {Execute it}
      fResult := GetCommState(Cid, DCB);
      csGetCommState := fResult;

      {Just return Cid for now}
      Long2Str(PChar(MsgBuf), Cid);
      SendMessage(MsgWin, wm_csGetCommState, StrLen(PChar(MsgBuf)), LongInt(MsgBuf));

      {re-patch}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

  function csSetCommBreak(Cid: Integer): Integer;
    {-Unpatch, call it, log it, repatch for next time}
  var
    fResult : Integer;
  begin
    with Patch[12] do begin
      Move(OrigCode, PatchPoint^, 5);

      {Execute it}
      fResult := SetCommBreak(Cid);
      csSetCommBreak := fResult;

      SendMessage(MsgWin, wm_csSetCommBreak, 0, 0);

      {re-patch}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

  function csSetCommEventMask(Cid: Integer; EvtMask: Word): PWord;
    {-Unpatch, call it, log it, repatch for next time}
  var
    fResult : PWord;
  begin
    with Patch[13] do begin
      Move(OrigCode, PatchPoint^, 5);

      {Execute it}
      fResult := SetCommEventMask(Cid, EvtMask);
      csSetCommEventMask := fResult;

      Long2Str(PChar(MsgBuf), EvtMask);
      SendMessage(MsgWin, wm_csSetCommEventMask, StrLen(PChar(MsgBuf)), LongInt(MsgBuf));

      {re-patch}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

  function csSetCommState(var DCB: TDCB): Integer;
    {-Unpatch, call it, log it, repatch for next time}
  var
    fResult : Integer;
  begin
    with Patch[14] do begin
      Move(OrigCode, PatchPoint^, 5);

      {Execute it}
      fResult := SetCommState(DCB);
      csSetCommState := fResult;

      {Just return fresult}
      Long2Str(PChar(MsgBuf), fResult);
      SendMessage(MsgWin, wm_csSetCommState, StrLen(PChar(MsgBuf)), LongInt(MsgBuf));

      {re-patch}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

  function csTransmitCommChar(Cid: Integer; Chr: Char): Integer;
    {-Unpatch, call it, log it, repatch for next time}
  var
    fResult : Integer;
  begin
    with Patch[15] do begin
      Move(OrigCode, PatchPoint^, 5);

      {Execute it}
      fResult := TransmitCommChar(Cid, Chr);
      csTransmitCommChar := fResult;

      {Return transmitted char}
      MsgBuf^[1] := Byte(Chr);
      MsgBuf^[2] := 0;
      SendMessage(MsgWin, wm_csTransmitCommChar, 1, LongInt(MsgBuf));

      {re-patch}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

  function csUngetCommChar(Cid: Integer; Chr: Char): Integer;
    {-Unpatch, call it, log it, repatch for next time}
  var
    fResult : Integer;
  begin
    with Patch[16] do begin
      Move(OrigCode, PatchPoint^, 5);

      {Execute it}
      fResult := UngetCommChar(Cid, Chr);
      csUngetCommChar := fResult;

      {Return "unget" char}
      MsgBuf^[1] := Byte(Chr);
      MsgBuf^[2] := 0;
      SendMessage(MsgWin, wm_csUngetCommChar, 1, LongInt(MsgBuf));

      {re-patch}
      Move(JmpCode, PatchPoint^, 5);
    end;
  end;

{Exported procedures for starting/stopping spying}

  procedure StartSpying(H : HWnd; Mode : Byte); export;
    {-Start spying, send messages to H}
  begin
    MsgWin := H;
    case Mode of
      smTotal   : PatchAll;
      smTrace,
      smCapture : PatchReadWrite;
      else        Exit;
    end;
    SpyMode := Mode;
  end;

  procedure StopSpying(H : HWnd); export;
    {-Stop spying}
  begin
    MsgWin := 0;
    case SpyMode of
      smTotal   : UnpatchAll;
      smTrace,
      smCapture : UnpatchReadWrite;
    end;
    SpyMode := smNone;
  end;

  Exports
    StartSpying index 1,
    StopSpying  index 2;

begin
  {Allocate message buffer}
  GetMem(MsgBuf, SizeOf(MsgBuf));
  if MsgBuf = nil then begin
    MessageBox(GetFocus, 'Out of memory', '', mb_IconExclamation+mb_Ok);
    Halt;
  end;
end.
